/*
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.facebook.presto.functionNamespace.mysql;

import com.facebook.airlift.json.JsonCodec;
import com.facebook.presto.spi.CatalogSchemaName;
import com.facebook.presto.spi.function.QualifiedFunctionName;
import com.facebook.presto.spi.function.RoutineCharacteristics;
import com.facebook.presto.spi.function.SqlInvokedFunction;
import com.facebook.presto.spi.function.SqlParameter;
import org.jdbi.v3.core.mapper.RowMapper;
import org.jdbi.v3.core.statement.StatementContext;

import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.List;
import java.util.Optional;

import static com.facebook.airlift.json.JsonCodec.jsonCodec;
import static com.facebook.airlift.json.JsonCodec.listJsonCodec;
import static com.facebook.presto.spi.type.TypeSignature.parseTypeSignature;

public class SqlInvokedFunctionRowMapper
        implements RowMapper<SqlInvokedFunction>
{
    private static final JsonCodec<List<SqlParameter>> SQL_PARAMETERS_CODEC = listJsonCodec(SqlParameter.class);
    private static final JsonCodec<RoutineCharacteristics> ROUTINE_CHARACTERISTICS_CODEC = jsonCodec(RoutineCharacteristics.class);

    @Override
    public SqlInvokedFunction map(ResultSet rs, StatementContext ctx)
            throws SQLException
    {
        String catalog = rs.getString("catalog_name");
        String schema = rs.getString("schema_name");
        String functionName = rs.getString("function_name");
        List<SqlParameter> parameters = SQL_PARAMETERS_CODEC.fromJson(rs.getString("parameters"));
        String returnType = rs.getString("return_type");
        String description = rs.getString("description");
        RoutineCharacteristics routineCharacteristics = ROUTINE_CHARACTERISTICS_CODEC.fromJson(rs.getString("routine_characteristics"));
        String body = rs.getString("body");
        long version = rs.getLong("version");

        return new SqlInvokedFunction(
                QualifiedFunctionName.of(new CatalogSchemaName(catalog, schema), functionName),
                parameters,
                parseTypeSignature(returnType),
                description,
                routineCharacteristics,
                body,
                Optional.of(version));
    }
}
